@@ -14,6 +14,7 @@ import dask
 import numpy as np
 import pandas as pd
 import xarray as xr
+from deprecation import deprecated
 from numpy import arange, array
 from scipy.sparse import coo_matrix
 from xarray import DataArray, Dataset
@@ -21,75 +22,190 @@ from xarray import DataArray, Dataset
 from linopy import expressions, variables
 from linopy.common import (
     _merge_inplace,
-    has_assigned_model,
+    dictsel,
+    forward_as_properties,
     has_optimized_model,
+    head_tail_range,
     is_constant,
+    maybe_replace_signs,
+    print_coord,
+    print_single_expression,
     replace_by_map,
 )
+from linopy.config import options
+from linopy.constants import EQUAL, GREATER_EQUAL, LESS_EQUAL
 
 
-class Constraint(DataArray):
+@forward_as_properties(
+    labels=[
+        "attrs",
+        "coords",
+        "indexes",
+        "name",
+        "shape",
+        "size",
+        "values",
+        "dims",
+        "ndim",
+    ],
+    lhs=["nterm"],
+)
+class Constraint:
     """
-    Constraint container for storing constraint labels.
+    Projection to a single constraint in a model.
 
     The Constraint class is a subclass of xr.DataArray hence most xarray
     functions can be applied to it.
     """
 
-    __slots__ = ("_cache", "_coords", "_indexes", "_name", "_variable", "model")
+    __slots__ = ("_model", "_name", "_labels")
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, labels: DataArray, model: Any, name: str):
+        """
+        Initialize the Constraint.
 
-        # workaround until https://github.com/pydata/xarray/pull/5984 is merged
-        if isinstance(args[0], DataArray):
-            da = args[0]
-            args = (da.data, da.coords)
-            kwargs.update({"attrs": da.attrs, "name": da.name})
+        Parameters
+        ----------
+        labels : xarray.DataArray
+            labels of the constraint.
+        model : linopy.Model
+            Underlying model.
+        name : str
+            Name of the constraint.
+        """
+        self._labels = labels
+        self._model = model
+        self._name = name
 
-        self.model = kwargs.pop("model", None)
-        super().__init__(*args, **kwargs)
-        assert self.name is not None, "Constraint data does not have a name."
+    @property
+    def labels(self):
+        """
+        Get the labels of the constraint.
+        """
+        return self._labels
 
-    # We have to set the _reduce_method to None, in order to overwrite basic
-    # reduction functions as `sum`. There might be a better solution (?).
-    _reduce_method = None
+    @property
+    def model(self):
+        """
+        Get the model of the constraint.
+        """
+        return self._model
 
-    # Disable array function, only function defined below are supported
-    # and set priority higher than pandas/xarray/numpy
-    __array_ufunc__ = None
-    __array_priority__ = 10000
+    @property
+    def name(self):
+        """
+        Get the name of the constraint.
+        """
+        return self._name
 
     def __repr__(self):
         """
-        Get the string representation of the constraints.
+        Get the string representation of the Constraint.
         """
-        data_string = (
-            "Constraint labels:\n" + self.to_array().__repr__().split("\n", 1)[1]
-        )
-        extend_line = "-" * len(self.name)
-        return (
-            f"Constraint '{self.name}':\n"
-            f"--------------{extend_line}\n\n"
-            f"{data_string}"
-        )
+        # return single if only one exist
+        if self.lhs.size == self.nterm:
+            expr_string = print_single_expression(
+                self.lhs.coeffs.values, self.lhs.vars.values, self.lhs.model
+            )
+            name_string = f" `{self.name}`" if self.name else ""
+            header_string = str(self.type) + name_string
+            header = f"{header_string}\n" + "-" * (len(header_string))
+            return f"{header}\n{expr_string} {self.sign.item()} {self.rhs.item()}"
 
-    def _repr_html_(self):
-        """
-        Get the html representation of the variables.
-        """
-        # return self.__repr__()
-        data_string = self.to_array()._repr_html_()
-        data_string = data_string.replace("xarray.DataArray", "linopy.Constraint")
-        return data_string
+        # create header string
+        name_string = f"`{self.name}`" if self.name else ""
+        shape_string = ", ".join(
+            [f"{self.dims[i]}: {self.shape[i]}" for i in range(self.ndim)]
+        )
+        shape_string = f"({shape_string})"
+        n_masked = (~self.mask).sum().item()
+        mask_string = f" - {n_masked} masked entries" if n_masked else ""
+        header_string = f"{self.type} {name_string} {shape_string}" + mask_string
+        header = f"{header_string}\n" + "-" * (len(header_string))
 
+        # create data string, print only a few values
+        max_print = options["display_max_rows"]
+        split_at = max_print // 2
+        to_print = np.flatnonzero(self.mask)
+        if not len(to_print):
+            # case that all entries are masked
+            to_print = head_tail_range(self.size, max_print)
+            data_string = "None"
+            return f"{header}\n{data_string}"
+
+        truncate = len(to_print) > max_print
+        if truncate:
+            to_print = np.hstack([to_print[:split_at], to_print[-split_at:]])
+
+        # create string, we use numpy to get the indexes
+        data_string = ""
+        idx = np.unravel_index(to_print, self.shape)
+        coords = [self.indexes[dim][idx[i]] for i, dim in enumerate(self.dims)]
+
+        # loop over all values to LinearExpression(print
+        coord_strings = []
+        expr_strings = []
+        sign_strings = []
+        rhs_strings = []
+        trunc_strings = []
+        for i in range(len(to_print)):
+            coord = {dim: c[i] for dim, c in zip(self.dims, coords)}
+            coord_string = print_coord(coord) + ":"
+            expr_string = print_single_expression(
+                self.coeffs.sel(coord).values,
+                self.vars.sel(coord).values,
+                self.lhs.model,
+            )
+            sign_string = f"{self.sign.sel(**dictsel(coord, self.sign.dims)).item()}"
+            rhs_string = f"{self.rhs.sel(**dictsel(coord, self.rhs.dims)).item()}"
+            trunc_string = "\n\t\t..." if i == split_at - 1 and truncate else ""
+
+            coord_strings.append(coord_string)
+            expr_strings.append(expr_string)
+            sign_strings.append(sign_string)
+            rhs_strings.append(rhs_string)
+            trunc_strings.append(trunc_string)
+
+        coord_width = max(len(c) for c in coord_strings)
+        expr_width = max(len(e) for e in expr_strings)
+
+        data_string = ""
+        for c, e, s, r, t in zip(
+            coord_strings, expr_strings, sign_strings, rhs_strings, trunc_strings
+        ):
+            data_string += f"\n{c:<{coord_width}} {e:<{expr_width}} {s:<2} {r}{t}"
+
+        return f"{header}{data_string}"
+
+    @deprecated(details="Use the `labels` property instead of `to_array`")
     def to_array(self):
         """
         Convert the variable array to a xarray.DataArray.
         """
-        return DataArray(self)
+        return self.labels
+
+    @property
+    def type(self):
+        """
+        Get the type of the constraint.
+        """
+        return "Constraint"
+
+    @property
+    def mask(self):
+        """
+        Get the mask of the constraint.
+
+        The mask indicates on which coordinates the constraint array is enabled
+        (True) and disabled (False).
+
+        Returns
+        -------
+        xr.DataArray
+        """
+        return (self.labels != -1).astype(bool)
 
     @property
-    @has_assigned_model
     def coeffs(self):
         """
         Get the left-hand-side coefficients of the constraint.
@@ -100,22 +216,14 @@ class Constraint(DataArray):
         return self.model.constraints.coeffs[self.name]
 
     @coeffs.setter
-    @has_assigned_model
     def coeffs(self, value):
         term_dim = self.name + "_term"
         value = DataArray(value).broadcast_like(self.vars, exclude=[term_dim])
-        try:
-            self.model.constraints.coeffs[self.name] = value
-        except ValueError:
-            coeffs = self.model.constraints.coeffs
-            coeffs = coeffs.assign_coords({term_dim: coeffs[term_dim]})
-            value = value.assign_coords({term_dim: value[term_dim]})
-            coeffs[self.name] = value
-            coeffs = coeffs.reset_index(term_dim, drop=True)
-            self.model.constraints.coeffs = coeffs
+        self.model.constraints.coeffs = self.model.constraints.coeffs.drop_vars(
+            self.name
+        ).assign({self.name: value})
 
     @property
-    @has_assigned_model
     def vars(self):
         """
         Get the left-hand-side variables of the constraint.
@@ -126,22 +234,18 @@ class Constraint(DataArray):
         return self.model.constraints.vars[self.name]
 
     @vars.setter
-    @has_assigned_model
     def vars(self, value):
         term_dim = self.name + "_term"
-        value = DataArray(value).broadcast_like(self.coeffs, exclude=[term_dim])
-        try:
-            self.model.constraints.vars[self.name] = value
-        except ValueError:
-            vars = self.model.constraints.vars
-            vars = vars.assign_coords({term_dim: vars[term_dim]})
-            value = value.assign_coords({term_dim: value[term_dim]})
-            vars[self.name] = value
-            vars = vars.reset_index(term_dim, drop=True)
-            self.model.constraints.vars = vars
+        if isinstance(value, variables.Variable):
+            value = value.labels
+        if not isinstance(value, DataArray):
+            raise TypeError("Expected value to be of type DataArray or Variable")
+        value = value.broadcast_like(self.coeffs, exclude=[term_dim])
+        self.model.constraints.vars = self.model.constraints.vars.drop_vars(
+            self.name
+        ).assign({self.name: value})
 
     @property
-    @has_assigned_model
     def lhs(self):
         """
         Get the left-hand-side linear expression of the constraint.
@@ -152,10 +256,10 @@ class Constraint(DataArray):
         term_dim = self.name + "_term"
         coeffs = self.coeffs.rename({term_dim: "_term"})
         vars = self.vars.rename({term_dim: "_term"})
-        return expressions.LinearExpression(Dataset({"coeffs": coeffs, "vars": vars}))
+        ds = Dataset({"coeffs": coeffs, "vars": vars})
+        return expressions.LinearExpression(ds, self.model)
 
     @lhs.setter
-    @has_assigned_model
     def lhs(self, value):
         if not isinstance(value, expressions.LinearExpression):
             raise TypeError("Assigned lhs must be a LinearExpression.")
@@ -165,7 +269,6 @@ class Constraint(DataArray):
         self.vars = value.vars
 
     @property
-    @has_assigned_model
     def sign(self):
         """
         Get the signs of the constraint.
@@ -176,16 +279,12 @@ class Constraint(DataArray):
         return self.model.constraints.sign[self.name]
 
     @sign.setter
-    @has_assigned_model
     @is_constant
     def sign(self, value):
-        value = DataArray(value).broadcast_like(self)
-        if (value == "==").any():
-            raise ValueError('Sign "==" not supported, use "=" instead.')
+        value = maybe_replace_signs(DataArray(value)).broadcast_like(self.sign)
         self.model.constraints.sign[self.name] = value
 
     @property
-    @has_assigned_model
     def rhs(self):
         """
         Get the right hand side constants of the constraint.
@@ -196,12 +295,11 @@ class Constraint(DataArray):
         return self.model.constraints.rhs[self.name]
 
     @rhs.setter
-    @has_assigned_model
     @is_constant
     def rhs(self, value):
         if isinstance(value, (variables.Variable, expressions.LinearExpression)):
             raise TypeError(f"Assigned rhs must be a constant, got {type(value)}).")
-        value = DataArray(value).broadcast_like(self)
+        value = DataArray(value).broadcast_like(self.rhs)
         self.model.constraints.rhs[self.name] = value
 
     @property
@@ -219,6 +317,10 @@ class Constraint(DataArray):
             )
         return self.model.dual[self.name]
 
+    @property
+    def shape(self):
+        return self.labels.shape
+
 
 @dataclass(repr=False)
 class Constraints:
@@ -268,7 +370,7 @@ class Constraints:
         self, names: Union[str, Sequence[str]]
     ) -> Union[Constraint, "Constraints"]:
         if isinstance(names, str):
-            return Constraint(self.labels[names], model=self.model)
+            return Constraint(self.labels[names], self.model, names)
 
         return self.__class__(
             self.labels[names],
@@ -288,7 +390,8 @@ class Constraints:
         """
         Provide method for the key-autocompletions in IPython.
 
-        See http://ipython.readthedocs.io/en/stable/config/integrating.html#tab-completion
+        See
+        http://ipython.readthedocs.io/en/stable/config/integrating.html#tab-completion
         For the details.
         """
         return list(self)
@@ -347,22 +450,23 @@ class Constraints:
         """
         Get the subset of constraints which are purely inequalities.
         """
-        return self[[n for n, s in self.sign.items() if s in ("<=", ">=")]]
+        return self[
+            [n for n, s in self.sign.items() if s in (LESS_EQUAL, GREATER_EQUAL)]
+        ]
 
     @property
     def equalities(self):
         """
         Get the subset of constraints which are purely equalities.
         """
-        return self[[n for n, s in self.sign.items() if s in ("=", "==")]]
+        return self[[n for n, s in self.sign.items() if s == EQUAL]]
 
     def sanitize_zeros(self):
         """
-        Filter out terms with zero coefficient.
+        Filter out terms with zero and close-to-zero coefficient.
         """
         for name in self:
-            term_dim = name + "_term"
-            not_zero = self.coeffs[name] != 0
+            not_zero = abs(self.coeffs[name]) > 1e-10
             self.vars[name] = self.vars[name].where(not_zero, -1)
             self.coeffs[name] = self.coeffs[name].where(not_zero)
 
@@ -565,62 +669,105 @@ class Constraints:
 
             shape = (ncons, nvars)  # same as model.nvars/ncons but already there
 
-        return coo_matrix((data, (rows, cols)), shape=shape)
+        # Group repeated variables in the same constraint
+        df = pd.DataFrame({"data": data, "rows": rows, "cols": cols})
+        df = df.groupby(["rows", "cols"], as_index=False).sum()
+
+        return coo_matrix((df.data, (df.rows, df.cols)), shape=shape)
 
 
+@dataclass(repr=False)
+@forward_as_properties(
+    labels=[
+        "attrs",
+        "coords",
+        "indexes",
+        "shape",
+        "size",
+        "values",
+        "dims",
+        "ndim",
+    ],
+    lhs=["nterm", "coeffs", "vars"],
+)
 class AnonymousConstraint:
     """
     A constraint container used for storing multiple constraint arrays.
     """
 
-    __slots__ = ("_lhs", "_sign", "_rhs")
+    _lhs: "expressions.LinearExpression"
+    _sign: Union[DataArray, str]
+    _rhs: Union[DataArray, float, int]
 
-    def __init__(self, lhs, sign, rhs):
+    def __post_init__(self):
         """
         Initialize a anonymous constraint.
         """
-        if isinstance(rhs, (variables.Variable, expressions.LinearExpression)):
-            raise TypeError(f"Assigned rhs must be a constant, got {type(rhs)}).")
-        self._lhs, self._rhs = xr.align(lhs, DataArray(rhs))
-        self._sign = DataArray(sign)
+        if isinstance(self.rhs, (variables.Variable, expressions.LinearExpression)):
+            raise TypeError(f"Assigned rhs must be a constant, got {type(self.rhs)}).")
+        lhs_data, rhs = xr.align(self.lhs.data, DataArray(self.rhs))
+        self._labels = (lhs_data.vars.chunk() + rhs).sum("_term")
+        self._labels.data = np.full(self.labels.shape, np.nan)
+        self._lhs = expressions.LinearExpression(lhs_data, self.lhs.model)
+        self._rhs = rhs
+        self._sign = DataArray(self.sign)
+
+    def __repr__(self):
+        """
+        Get the representation of the AnonymousConstraint.
+        """
+        return Constraint.__repr__(self)
+
+    @property
+    def name(self):
+        return ""
+
+    @property
+    def type(self):
+        """
+        Get the type of the constraint.
+        """
+        # needed for __repr__ compatibility.
+        return "AnomymousConstraint"
+
+    @property
+    def mask(self):
+        """
+        Get the mask of the constraint.
+        """
+        # needed for __repr__ compatibility.
+        return xr.full_like(self.labels, True, dtype=bool)
+
+    @property
+    def labels(self):
+        """
+        Get the labels of the constraint.
+        """
+        return self._labels
 
     @property
     def lhs(self):
+        """
+        Get the left hand side of the constraint.
+        """
         return self._lhs
 
     @property
     def sign(self):
+        """
+        Get the sign of the constraint.
+        """
         return self._sign
 
     @property
     def rhs(self):
+        """
+        Get the right hand side of the constraint.
+        """
         return self._rhs
 
-    def __repr__(self):
-        """
-        Get the string representation of the expression.
-        """
-        lhs_string = self.lhs.to_dataset().__repr__()  # .split("\n", 1)[1]
-        lhs_string = lhs_string.split("Data variables:\n", 1)[1]
-        lhs_string = lhs_string.replace("    coeffs", "coeffs")
-        lhs_string = lhs_string.replace("    vars", "vars")
-        if self.sign.size == 1:
-            sign_string = self.sign.item()
-        else:
-            sign_string = self.sign.__repr__().split("\n", 1)[1]
-        if self.rhs.size == 1:
-            rhs_string = self.rhs.item()
-        else:
-            rhs_string = self.rhs.__repr__().split("\n", 1)[1]
-        return (
-            f"Anonymous Constraint:\n"
-            f"---------------------\n"
-            f"\n{lhs_string}"
-            f"\n{sign_string}"
-            f"\n{rhs_string}"
-        )
-
-    def from_rule(model, rule, coords):
+    @classmethod
+    def from_rule(cls, model, rule, coords):
         """
         Create a constraint from a rule and a set of coordinates.
 
@@ -676,18 +823,19 @@ class AnonymousConstraint:
             raise TypeError(msg)
 
         combinations = product(*[c.values for c in coords.values()])
-        placeholder_lhs = expressions.ScalarLinearExpression((np.nan,), (-1,))
+        placeholder_lhs = expressions.ScalarLinearExpression((np.nan,), (-1,), model)
         placeholder = AnonymousScalarConstraint(placeholder_lhs, "=", np.nan)
         cons = [rule(model, *coord) or placeholder for coord in combinations]
         exprs = [con.lhs for con in cons]
 
-        lhs = expressions.LinearExpression._from_scalarexpression_list(exprs, coords)
+        lhs = expressions.LinearExpression._from_scalarexpression_list(
+            exprs, coords, model
+        )
         sign = DataArray(array([c.sign for c in cons]).reshape(shape), coords)
         rhs = DataArray(array([c.rhs for c in cons]).reshape(shape), coords)
-        return AnonymousConstraint(lhs, sign, rhs)
+        return cls(lhs, sign, rhs)
 
 
-@dataclass
 class AnonymousScalarConstraint:
     """
     Container for anonymous scalar constraint.
@@ -696,9 +844,49 @@ class AnonymousScalarConstraint:
     (rhs) for exactly one constraint.
     """
 
-    lhs: expressions.ScalarLinearExpression
-    sign: str
-    rhs: float
+    _lhs: "expressions.ScalarLinearExpression"
+    _sign: str
+    _rhs: float
+
+    def __init__(self, lhs, sign, rhs):
+        """
+        Initialize a anonymous scalar constraint.
+        """
+        if not isinstance(rhs, (int, float, np.float32, np.float64, np.integer)):
+            raise TypeError(f"Assigned rhs must be a constant, got {type(rhs)}).")
+        self._lhs = lhs
+        self._sign = sign
+        self._rhs = rhs
+
+    def __repr__(self):
+        """
+        Get the representation of the AnonymousScalarConstraint.
+        """
+        expr_string = print_single_expression(
+            self.lhs.coeffs, self.lhs.vars, self.lhs.model
+        )
+        return f"AnonymousScalarConstraint: {expr_string} {self.sign} {self.rhs}"
+
+    @property
+    def lhs(self):
+        """
+        Get the left hand side of the constraint.
+        """
+        return self._lhs
+
+    @property
+    def sign(self):
+        """
+        Get the sign of the constraint.
+        """
+        return self._sign
+
+    @property
+    def rhs(self):
+        """
+        Get the right hand side of the constraint.
+        """
+        return self._rhs
 
     def to_anonymous_constraint(self):
         return AnonymousConstraint(self.lhs.to_linexpr(), self.sign, self.rhs)
